# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from concurrent.futures import CancelledError
from typing import TYPE_CHECKING

from google.cloud.pubsub_v1 import SubscriberClient # type: ignore
from google.cloud.pubsub_v1.subscriber.futures import StreamingPullFuture # type: ignore

from aorta.pool import ThreadWorker
if TYPE_CHECKING:
    from ._topiclistener import TopicListener


class PubsubWorker(ThreadWorker):
    streamer: StreamingPullFuture
    pool: 'TopicListener'

    def __init__(
        self,
        pool: 'TopicListener',
        prefetch: int = 1
    ):
        super().__init__(pool)
        self.prefetch = prefetch

    def beat(self) -> None:
        if not self.must_exit:
            try:
                self.streamer.result(timeout=0.5)
            except CancelledError:
                self.logger.warning("Broker connection stream was cancelled.")
            except TimeoutError:
                pass
            except Exception as e:
                self.logger.critical(
                    "Caught fatal %s on broker connection.",
                    type(e).__name__
                )
                self.must_exit = True

        if self.must_exit and self.streamer.running():
            self.logger.debug("Cancelling leases with Google Pubsub")
            self.streamer.cancel()

    def configure(self) -> None:
        super().configure()
        self.subscription_id = os.environ['GOOGLE_PUBSUB_SUBSCRIPTION']
        self.subscriber = SubscriberClient()
        self.streamer = self.subscriber.subscribe( # type: ignore
            self.subscription_id,
            self.accept_message,
            flow_control=self.pool.get_flow_control()
        )

        self.logger.debug("Established connection to subscription %s", self.subscription_id)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import contextlib
import logging
import os
import threading
import time
from typing import Any
from typing import Coroutine
from typing import Iterable

from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IMessageHandler
from aorta.types import IPool
from aorta.provider import Provider


class Worker(threading.Thread):
    interval: float = 0.1
    logger: logging.Logger = logging.getLogger('aorta.transport')
    must_exit: bool = False
    tasks: set[asyncio.Task[None]]

    def __init__(self, pool: IPool):
        self.pool = pool
        self.worker_id = bytes.hex(os.urandom(24))
        super().__init__(
            target=lambda: asyncio.run(self.main_event_loop()),
            daemon=True
        )
        self.start()

    def accept_message(self, frame: Acknowledgable):
        self.pool.notify(frame)
        if self.must_exit:
            # Stop accepting messages if we must stop.
            self.logger.debug(
                "Message rejected because the listener is shutting down (id: %s)",
                frame.message_id # type: ignore
            )
            self.create_task(self.reject(frame))
            return
        envelope = None
        try:
            envelope = Provider.loads(frame.data)
            if envelope is None:
                raise ValueError("Unable to parse message from incoming data.")
        except Exception as e:
            self.create_task(self.acknowledge(frame))
            self.logger.warning("Caught fatal %s", repr(e))
            return
        assert envelope is not None
        if isinstance(envelope, Envelope):
            self.on_message_received(frame, envelope) # type: ignore
        else:
            self.create_task(self.reject(frame))
            self.logger.warning(
                "Received unknown Aorta message (kind: %s, uid: %s, id: %s)",
                envelope.kind,
                envelope.metadata.uid[:6],
                envelope.metadata.correlation_id[:6]
            )

    def create_task(self, c: Coroutine[Any, Any, Any]):
        self.tasks.add(self.loop.create_task(c))

    def configure(self) -> None:
        self.loop = asyncio.get_event_loop()
        self.tasks = set()

    @contextlib.contextmanager
    def event_loop(self, envelope: Envelope[Any]):
        try:
            yield self.loop
        except Exception as e:
            self.logger.exception(
                "Caught fatal %s while handling message (uid: %s, correlation-id: %s)",
                type(e).__name__,
                envelope.metadata.uid,
                envelope.metadata.correlation_id
            )
            return True

    def join(self, timeout: float | None = None):
        self.stop()

    def on_message_received(
        self,
        frame: Acknowledgable,
        envelope: Envelope[Any],
    ):
        self.create_task(self.execute(frame, envelope, Provider.get(envelope)))
        
    def on_task_completed(
        self,
        task: asyncio.Task[Any],
        envelope: Envelope[Any],
        handlers: Iterable[type[IMessageHandler]]
    ) -> None:
        pass

    def stop(self):
        self.must_exit = True
        self.logger.info("Tearing down worker (id: %s)", self.ident)
        return self

    async def acknowledge(self, frame: Acknowledgable):
        self.loop.run_in_executor(None, frame.ack)

    async def reject(self, frame: Acknowledgable):
        self.loop.run_in_executor(None, frame.nack)

    async def main_event_loop(self):
        self.logger.debug("Spawning worker (thread: %s)", self.ident)
        self.publisher = self.pool.get_publisher()
        self.runner = self.pool.get_runner(self.publisher)
        self.configure()
        self.logger.debug("Start polling for new messages (thread: %s)", self.ident)
        while True:
            # Check pending tasks before doing anything.
            if self.tasks:
                done, *_ = await asyncio.wait(
                    fs=self.tasks,
                    timeout=0.1,
                    return_when='FIRST_COMPLETED'
                )
                for t in done:
                    self.tasks.remove(t)
            if self.must_exit:
                if self.tasks:
                    await asyncio.wait(self.tasks)
                break
            time.sleep(self.interval)

    async def execute(
        self,
        frame: Acknowledgable,
        envelope: Envelope[Any],
        handlers: Iterable[type[IMessageHandler]]
    ) -> None:
        try:
            return await self.runner.run(
                publisher=self.publisher,
                envelope=envelope,
                handlers=handlers
            )
        #except Exception:
        #    if len(list(handlers)) <= 1:
        #        await self.reject(frame)
        #    raise
        finally:
            await self.acknowledge(frame)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


class Drop(Exception):
    """Instructs the handler to drop a message."""
    __module__: str = 'aorta.types'
    reason: str

    def __init__(self, reason: str = ''):
        self.reason = reason


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import logging.config
import time
from typing import Any
from typing import TypeVar

from libcanonical.runtime import MainProcess # type: ignore

from aorta import Command
from aorta import Event
from aorta import MessagePublisher
from aorta import NullTransport
from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IPublisher
from aorta.types import IRunner
from aorta.pool import ThreadWorker

E = TypeVar('E', bound=Command|Event)


class BaseConsumer(MainProcess):
    __module__: str = 'aorta.ext.consumer'
    concurrency: int = 1
    debug: bool = False
    interval = 0.25
    logger: logging.Logger = logging.getLogger('aorta.transport.ingress')
    max_messages: int = 0
    metrics: logging.Logger = logging.getLogger('aorta.metrics')
    num_workers: int = 1
    workers: list[ThreadWorker]

    def __init__(
        self,
        name: str,
        workers: int = 1,
        concurrency: int = 1,
        max_runtime: float = 0.0,
        max_messages: int = 0,
        loglevel: str = 'INFO',
        transport_loglevel: str = 'WARNING',
        **kwargs: Any
    ):
        super().__init__(name=name)
        self.concurrency = concurrency
        self.loglevel = loglevel
        self.max_runtime = max_runtime
        self.max_messages = max_messages
        self.max_workers = workers
        self.received = 0
        self.transport_loglevel = transport_loglevel
        self.workers = []

    def configure(self, reloading: bool = False):
        self.logger.debug("Running with %s workers", self.concurrency)
        if not reloading:
            self.publisher = self.get_publisher()
            for _ in range(self.max_workers):
                self.workers.append(self.initialize_worker())
        if self.max_messages:
            self.logger.info("Worker will consume at most %s messages", self.max_messages)

    def configure_worker(self) -> None:
        logging.config.dictConfig(dict(self.get_logging_config()))

    def get_publisher(self) -> IPublisher:
        return MessagePublisher(
            transport=NullTransport()
        )

    def get_runner(self, publisher: IPublisher) -> IRunner:
        raise NotImplementedError

    def initialize_worker(self) -> ThreadWorker:
        raise NotImplementedError

    def notify(self, frame: Acknowledgable):
        self.received += 1

    def route(self, envelope: Envelope[E]) -> list[str]:
        """Return a list of strings indicating the topics that the
        `envelope` must be sent to. Subclasses must override this
        method.
        """
        raise NotImplementedError

    def main_event(self) -> None:
        if self.max_messages:
            must_exit = any([
                self.received >= self.max_messages,
                self.max_runtime and (self.age > self.max_runtime) and self.received == 0
            ])
            if not must_exit:
                return

            self.logger.warning(
                "Received maximum amount of %s messages after %0.2fs, exiting.",
                self.received,
                self.age
            )
            self.stop()
            for worker in self.workers:
                worker.stop()
                self.logger.debug("Closing worker (id: %s)", worker.ident)
            t0 = time.monotonic()
            while any([w.is_alive() for w in self.workers]):
                time.sleep(1)
                dt = (time.monotonic() - t0)
                if dt > 180:
                    self.logger.warning("Workers did not stop after %.02f seconds.", dt)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .acknowledgable import Acknowledgable
from .ipublisher import IPublisher
from .irunner import IRunner


class IPool(Protocol):
    __module__: str = 'aorta.types'

    def get_publisher(self) -> IPublisher:
        ...

    def get_runner(self, publisher: IPublisher) -> IRunner:
        ...

    def notify(self, frame: Acknowledgable) -> None:
        ...


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

import pytest

import aorta
from aorta.types.test.messagetype import *
from aorta.types.test.messagehandler import *
from aorta.types.test.transaction import *
from ..conftest import FooEvent


@pytest.fixture # type: ignore
def typecheck() -> Callable[[aorta.types.MessageHeader], bool]:
    return aorta.types.MessageHeader.is_event


@pytest.fixture # type: ignore
def parse() -> Callable[[Any], aorta.types.Envelope[Any] | aorta.types.MessageHeader | None]:
    return aorta.parse


@pytest.fixture # type: ignore
def MessageHandler(EventListener: type[aorta.MessageHandler[Any]]) -> type[aorta.MessageHandler[Any]]:
    return EventListener


@pytest.fixture
def message() -> aorta.types.Publishable:
    return FooEvent(foo=1)